\documentclass[11pt,a4paper]{article}
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc} % better fonts
\usepackage{hyperref}
\usepackage[left=2cm,right=2cm,bottom=3cm,top=1cm]{geometry}

\usepackage{color}
\definecolor{lightgrey}{gray}{0.35}
\definecolor{darkgrey}{gray}{0.20}

% code listings in our own style
\usepackage{listings}
\definecolor{lightestyellow}{rgb}{1,1,0.93}
\definecolor{dkgreen}{rgb}{0,.2,0}
\definecolor{dkblue}{rgb}{0,0,.3}
\definecolor{dkyellow}{cmyk}{0,0,.7,.5}
\definecolor{gray}{gray}{0.50}
\lstset{ language = Haskell,		basicstyle   = \scriptsize\ttfamily,
  keywordstyle    = \color{dkblue},	stringstyle  = \color{red},
  identifierstyle = \color{dkgreen},	backgroundcolor = \color{lightestyellow},
  rulecolor       = \color{gray},	showtabs        = false,
  tabsize         = 8,			breaklines      = true,
  frame           = single,		showstringspaces= false
}
\lstnewenvironment{code}[0]{}{}
\lstnewenvironment{bashCode}[0]{\lstset{language=bash}}{}

\title{Finding Cheryl's birthday with DEMO}
\author{Malvin Gattinger}
\date{April 2015}

\begin{document}

\maketitle

\section{Source Code}

Download \texttt{DEMO\_S5.hs} and \texttt{EREL.hs} from \url{http://homepages.cwi.nl/~jve/software/demo_s5} and put them in the same directory. Then you can use \texttt{runhaskell cheryl.lhs}.

\begin{code}
module Main where

import DEMO_S5
import Data.List
import Data.Function

possibilities :: [(Int, String)]
possibilities = [ (15,"May"), (16,"May"), (19,"May"),
  (17,"June"), (18,"June"), (14,"July"), (16,"July"),
  (14,"August"), (15,"August"), (17,"August") ]

init_Cheryl :: EpistM (Int,String)
init_Cheryl = Mo possibilities [a,b] [] rels possibilities where
  rels = [ ( a, groupBy ((==) `on` snd) possibilities ),
	   ( b, groupBy ((==) `on` fst) (sortBy (compare `on` fst) possibilities) ) ]

knWhich :: Agent -> Form (Int, [Char])
knWhich i = Disj [ Kn i (Info s) | s <- possibilities ]

listWorlds :: EpistM (Int,String) -> IO ()
listWorlds (Mo _ _ _ _ actuals) = putStrLn (" --> " ++ (show actuals))

main :: IO ()
main = do
  putStrLn "We start with all possibilities."
  listWorlds init_Cheryl
  putStrLn "Albert: I don't know when Cheryl's birthday is and I know that Bernard does not know."
  let model2 = (upd_pa init_Cheryl (Conj [Ng $ knWhich a, Kn a $ Ng (knWhich b)]))
  listWorlds model2
  putStrLn "Bernard: Now I know when Cheryl's birthday is."
  let model3 = (upd_pa model2 (knWhich b))
  listWorlds model3
  putStrLn "Albert: Now I also know when Cheryl's birthday is."
  let model4 = (upd_pa model3 (knWhich a))
  listWorlds model4
  putStrLn "Bye bye."
\end{code}

\section{Output}

\begin{bashCode}
We start with all possibilities.
 --> [(15,"May"),(16,"May"),(19,"May"),(17,"June"),(18,"June"),(14,"July"),(16,"July"),(14,"August"),(15,"August"),(17,"August")]
Albert: I don't know when Cheryl's birthday is and I know that Bernard does not know.
 --> [(14,"July"),(16,"July"),(14,"August"),(15,"August"),(17,"August")]
Bernard: Now I know when Cheryl's birthday is.
 --> [(16,"July"),(15,"August"),(17,"August")]
Albert: Now I also know when Cheryl's birthday is.
 --> [(16,"July")]
Bye bye.
\end{bashCode}

\end{document}
